from numpy import dtype, ndarray
from numpy import float32 as f32
from numpy import float64 as f64

from rlic._typing import BoundaryPair, UVMode

def convolve_f32(
    texture: ndarray[tuple[int, int], dtype[f32]],
    uv: tuple[
        ndarray[tuple[int, int], dtype[f32]],
        ndarray[tuple[int, int], dtype[f32]],
        UVMode,
    ],
    kernel: ndarray[tuple[int], dtype[f32]],
    boundaries: tuple[BoundaryPair, BoundaryPair],
    iterations: int = 1,
) -> ndarray[tuple[int, int], dtype[f32]]: ...
def convolve_f64(
    texture: ndarray[tuple[int, int], dtype[f64]],
    uv: tuple[
        ndarray[tuple[int, int], dtype[f64]],
        ndarray[tuple[int, int], dtype[f64]],
        UVMode,
    ],
    kernel: ndarray[tuple[int], dtype[f64]],
    boundaries: tuple[BoundaryPair, BoundaryPair],
    iterations: int = 1,
) -> ndarray[tuple[int, int], dtype[f64]]: ...



from dataclasses import replace
from itertools import permutations, product

import pytest
from pytest import RaisesExc, RaisesGroup

from rlic._boundaries import (
    COMBO_ALLOWED_BOUNDS,
    COMBO_DISALLOWED_BOUNDS,
    SUPPORTED_BOUNDS,
    BoundarySet,
)


@pytest.mark.parametrize(
    "bounds_input, expected_output",
    [
        pytest.param("a", BoundarySet(x=("a", "a"), y=("a", "a")), id="expand-all"),
        pytest.param(
            {"x": "a", "y": "b"},
            BoundarySet(x=("a", "a"), y=("b", "b")),
            id="expand-keys",
        ),
        pytest.param(
            {"x": ("a", "b"), "y": ("c", "w")},
            BoundarySet(x=("a", "b"), y=("c", "w")),
            id="already-expanded",
        ),
        pytest.param(
            {"x": ["a", "b"], "y": ["c", "w"]},
            BoundarySet(x=("a", "b"), y=("c", "w")),
            id="lists-to-tuples",
        ),
    ],
)
def test_from_user_input(bounds_input, expected_output):
    result = BoundarySet.from_user_input(bounds_input)
    assert result == expected_output


@pytest.mark.parametrize(
    "invalid_input",
    [
        123,
        None,
        ["a", "b"],  # List is not a valid input type
        {"x": "a"},  # Missing 'y' key
        {"y": "b"},  # Missing 'x' key
        {"x": "a", "y": "b", "z": "c"},  # extra key not allowed
    ],
)
def test_from_user_input_invalid_type(invalid_input):
    assert BoundarySet.from_user_input(invalid_input) is None


@pytest.mark.parametrize(
    "bound_input",
    [
        pytest.param(BoundarySet(x=(b1, b1), y=(b2, b2)), id=f"doubles_x={b1}_y={b2}")
        for (b1, b2) in product(SUPPORTED_BOUNDS, SUPPORTED_BOUNDS)
    ],
)
def test_validate_simple_bounds(bound_input):
    assert not bound_input.collect_exceptions()
    assert bound_input.validate() is None


@pytest.mark.parametrize(
    "bound_input",
    [
        pytest.param(BoundarySet(x=(b1, b1), y=(b2, b2)), id=f"doubles_x={b1}_y={b2}")
        for (b1, b2) in product(COMBO_ALLOWED_BOUNDS, COMBO_ALLOWED_BOUNDS)
    ],
)
def test_validate_bounds_combos(bound_input):
    assert not bound_input.collect_exceptions()
    assert bound_input.validate() is None


@pytest.mark.parametrize(
    "bound_input",
    [
        pytest.param(BoundarySet(x=(b1, b2), y=(b1, b1)), id=f"{b1}+{b2}")
        for (b1, b2) in product(COMBO_DISALLOWED_BOUNDS, SUPPORTED_BOUNDS)
        if b1 != b2
    ],
)
def test_invalid_combos(bound_input):
    with pytest.raises(ValueError, match=r"^left x boundary '\w+' cannot be combined"):
        bound_input.validate()

    reversed_x = replace(bound_input, x=reversed(bound_input.x))
    with pytest.raises(ValueError, match=r"^right x boundary '\w+' cannot be combined"):
        reversed_x.validate()

    swapped = BoundarySet(x=bound_input.y, y=bound_input.x)
    with pytest.raises(ValueError, match=r"^left y boundary '\w+' cannot be combined"):
        swapped.validate()

    swapped_reversed_y = replace(swapped, y=reversed(swapped.y))
    with pytest.raises(ValueError, match=r"^right y boundary '\w+' cannot be combined"):
        swapped_reversed_y.validate()


@pytest.mark.parametrize(
    "bound_input",
    [
        pytest.param(
            BoundarySet(x=(b1, b2), y=(b3, b4)),
            id=f"x=({b1}, {b2}), y=({b3}, {b4})",
        )
        for (b1, b2, b3, b4) in permutations(
            ("periodic", "periodic", "periodic", "unknown")
        )
    ],
)
def test_unknown_bounds(bound_input):
    with pytest.raises(
        ValueError, match=r"^Unknown (left|right) (x|y) boundary 'unknown'$"
    ):
        bound_input.validate()


def test_multiple_errors():
    with RaisesGroup(
        RaisesExc(ValueError, match=r"^Unknown left x boundary 'unknown'$"),
        RaisesExc(ValueError, match=r"^left y boundary 'periodic' cannot be combined"),
        match="Found multiple issues with boundary specifications",
    ):
        BoundarySet(x=("unknown", "periodic"), y=("periodic", "closed")).validate()



from __future__ import annotations

__all__ = [
    "COMBO_ALLOWED_BOUNDS",
    "COMBO_DISALLOWED_BOUNDS",
    "SUPPORTED_BOUNDS",
    "BoundarySet",
]

import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from rlic._typing import AnyBoundary, Boundary, BoundaryDict, BoundaryPair
if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup  # pyright: ignore[reportUnreachable]


# boundaries that can be combined with another value on the opposite side
COMBO_ALLOWED_BOUNDS = frozenset({"closed"})
# boundaries that require the exact same value be used on the opposite side
COMBO_DISALLOWED_BOUNDS = frozenset({"periodic"})

SUPPORTED_BOUNDS = frozenset(COMBO_ALLOWED_BOUNDS | COMBO_DISALLOWED_BOUNDS)


def as_pair(b: AnyBoundary, /) -> BoundaryPair:
    match b:
        case str():
            return (b, b)
        case (str(), str()):
            return (b[0], b[1])
        case _:  # pragma: no cover # pyright: ignore[reportUnnecessaryComparison]
            raise RuntimeError  # pyright: ignore[reportUnreachable]


@dataclass(frozen=True, slots=True, kw_only=True)
class BoundarySet:
    x: BoundaryPair
    y: BoundaryPair

    @staticmethod
    def from_user_input(bounds: Boundary | BoundaryDict, /) -> BoundarySet | None:
        # this function is responsible for validating the keys in a dict input, but
        # lets through any str as keys. It should output a very predictable structure.
        match bounds:
            case str():
                return BoundarySet(x=as_pair(bounds), y=as_pair(bounds))
            case {"x": str() | (str(), str()), "y": str() | (str(), str())} if (
                len(bounds) == 2
            ):
                return BoundarySet(x=as_pair(bounds["x"]), y=as_pair(bounds["y"]))
            case _:
                # signal an invalid input
                return None

    def collect_exceptions(self) -> list[Exception]:
        # this function is responsible for invalidating unsupported keys (or combos)
        msg_unknown = "Unknown {side} {ax} boundary {name!r}"
        msg_invalid_combo = (
            "{side} {ax} boundary {name!r} cannot be combined with "
            "a different boundary ({other!r})"
        )
        exceptions: list[Exception] = []

        for axis, (left, right) in [("x", self.x), ("y", self.y)]:
            if {left, right}.issubset(COMBO_ALLOWED_BOUNDS):
                continue

            if {left, right}.issubset(SUPPORTED_BOUNDS):
                # intentionally only report on disallowed combos
                # if both bounds are known (avoid over-reporting for simple typos)
                if left == right:
                    continue
                if left in COMBO_DISALLOWED_BOUNDS:
                    msg = msg_invalid_combo.format(
                        side="left", name=left, other=right, ax=axis
                    )
                    exceptions.append(ValueError(msg))
                if right in COMBO_DISALLOWED_BOUNDS:
                    msg = msg_invalid_combo.format(
                        side="right", name=right, other=left, ax=axis
                    )
                    exceptions.append(ValueError(msg))
            else:
                if left not in SUPPORTED_BOUNDS:
                    msg = msg_unknown.format(side="left", name=left, ax=axis)
                    exceptions.append(ValueError(msg))
                if right not in SUPPORTED_BOUNDS:
                    msg = msg_unknown.format(side="right", name=right, ax=axis)
                    exceptions.append(ValueError(msg))

        return exceptions

    def validate(self) -> None:
        exceptions = self.collect_exceptions()
        if len(exceptions) == 1:
            raise exceptions[0]
        elif exceptions:
            raise ExceptionGroup(
                "Found multiple issues with boundary specifications", exceptions
            )



__all__ = [
    "Boundary",
    "UVMode",
]

from typing import Literal, TypeAlias, TypedDict

Boundary = Literal["closed", "periodic"]
BoundaryPair = tuple[Boundary, Boundary]
AnyBoundary: TypeAlias = Boundary | BoundaryPair


class BoundaryDict(TypedDict):
    x: AnyBoundary
    y: AnyBoundary


UVMode = Literal["velocity", "polarization"]
